#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include "codeGen/RegisterAllocator.h"
#include <filesystem>
#include <iostream>
#include <cassert>

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	return;
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
	    ".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (const auto& c : scope->classes) {
		c->calculateSize();
	}

	for (const auto& f : scope->functions) {
		compileFunction(f);
	}

	return buf.str();
}

void Compiler::compileFunction(const FunctionInfoPtr& f) {
	string name = mangleFunctionName(f);
	buf << ".globl \"" << name << "\"\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << "\"" << name << "\":\n";
	QuadrupleGenerator::Result quadEnv = quadGen.compileFunction(f);
	try {
		RegisterAllocator regGen((quadEnv));
		regGen.allocate();
	} catch(const ParseError &e) {
		printFunction(quadEnv);
		throw ParseError(buf.str() + e.what());
	}

	printFunction(quadEnv);
}

void Compiler::printFunction(QuadrupleGenerator::Result quadEnv) {
	int blkNo = 0;
	for (const auto& b : quadEnv.blocks) {
		buf << "----------------------------------------------------------------------------\n";
		buf << "blok " << blkNo << " " << b->getName() << endl;
		/*buf << "in: ";
		for (auto in : b->in) {
			buf << in->getName() << " ";
		}
		buf << "\nouts: ";
		for (auto out : b->out) {
			buf << out->getName() << " ";
		}
		buf << "\n";*/
		if (!b->flow.in.empty()) {
			buf << "in: ";
			for (const auto& v : b->flow.in) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		if (!b->flow.def.empty()) {
			buf << "def: ";
			for (const auto& v : b->flow.def) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		for (const auto& q : b->quads) {
			auto ret = q->toString();
			if (!ret.empty()) {
				auto len = 0;
				for (const auto &c : ret) {
					len++;
					if (c == '\t') len += 8 - (len % 8);
				}
				for (; len < 70; len++) ret += ' ';
				buf << ret << " alive: ";
				for (auto var : q->aliveAfter)
					buf << var->name << " ";
				buf << endl;
			}
		}
		if (!b->flow.use.empty()) {
			buf << "----------\nuse: ";
			for (const auto &v : b->flow.use) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		if (!b->flow.out.empty()) {
			buf << "----------\nout: ";
			for (const auto &v : b->flow.out) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		blkNo++;
	}
}

Compiler::regs = {"eax", "ebx", "ecx", "edx", "esi", "edi"};

string Compiler::getRef(VariablePtr &v) {
	if (v->registerColor == -1) {
		if (!v->localOffset) {
			temps.emplace(v);
			v->localOffset = 4 * temps.size();
		}
		return to_string(v->localOffset) + "(ebp)";
	} else {
		return regs[v->registerColor+1];
	}
}

void Compiler::generateQJumpCond(QJumpCond &q) {
//	return QJump::generateAsm();

}

void Compiler::generateQAlloc(QAlloc &q) {


}

void Compiler::generateQWrite(QWrite &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQAccess(QAccess &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQReturn(QReturn &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQCall(QCall &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQParam(QParam &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQJump(QJump &q) {
//	return Quadruple::generateAsm();

}

void Compiler::generateQAssign(QAssign &q) {
	string tg = regs[q.loc->registerColor + 1];
	if (q.args.size() == 1) {
		switch (q.op.op) {
			case Op::Not: {
				buf << ""
				return;
			}
		}
	} else {
		buf << "movl %" << tg << " ";
		auto &i = q.args[0], &j = q.args[1];
		buf << getRef(i) << endl;
		buf << getOpName(q.op) << " %" << tg << " " << getRef(i) << endl;
	}
}

void Compiler::generateQLabel(QLabel &q) {
//	return Quadruple::generateAsm();

}