#ifndef COMPILER_HEADER
#define COMPILER_HEADER

#include "TypeDefs.h"
#include "Absyn.h"
#include "Skeleton.h"
#include <filesystem>
#include <utility>
#include "codeGen/Variable.h"
#include "ParseError.h"
#include "Info.h"
#include "codeGen/Quadruple.h"
#include "codeGen/BasicBlock.h"
#include "codeGen/QuadrupleGenerator.h"
#include <unordered_set>

using Op = Quadruple::Op;
class Compiler {
public:
	Compiler(const std::string& f, shared_ptr<Scope> s) : file(f), buf(), scope(s), quadGen(std::move(s)) {};

	static const std::string extension;
	static void externalCommand(std::filesystem::path file);

	string compile(Visitable *v);

	static std::string getVirtName(const ClassInfoPtr& c) {
		return c->name + ":virt_table";
	}

	void generateQJumpCond(QJumpCond &q);
	void generateQAlloc(QAlloc &q);
	void generateQWrite(QWrite &q);
	void generateQAccess(QAccess &q);
	void generateQReturn(QReturn &q);
	void generateQCall(QCall &q);
	void generateQParam(QParam &q);
	void generateQJump(QJump &q);
	void generateQAssign(QAssign &q);
	void generateQLabel(QLabel &q);
private:
	std::filesystem::path file;
	std::stringstream buf;
	shared_ptr<Scope> scope;
	QuadrupleGenerator quadGen;

	unordered_set<VariablePtr> temps;
	unordered_set<VarInfoPtr> localInfos;

	void compileFunction(const FunctionInfoPtr& f);
	void printFunction(QuadrupleGenerator::Result quadEnv);

	static std::string mangleFunctionName(const FunctionInfoPtr& f) {
		if (auto c = f->klass.lock()) {
			return c->name + "::" + f->name;
		}
		return f->name;
	}
	string getRef(VariablePtr &v);


	static const char** regs;

	static string getOpName(Quadruple::Op op) {
		switch(op.op) {
			case Op::Plus: return "addl";
			case Op::Minus: return "subl";
			case Op::Mul: return "mull";
			case Op::Div: return "divl";
			case Op::Mod: return "divl";
			default: return "";
		}
	}

	string moveToAnyRegister(VariablePtr &v) {
		if (v->registerColor == -1) {
			buf << "movl %eax " << getRef(v) << endl;
			return "eax";
		} else return regs[v->registerColor+1];
	}
};

#endif
